#include <cstdio>
#include <cstdarg>
#include <iostream>
#ifdef _WIN32
#include <windows.h>
#else
#include <sys/ioctl.h>
#include <unistd.h>
#endif
#include "textoutput.h"

char Textout::text[MAX_BUF_SIZE];
int Textout::curLine = 0;
int Textout::bufferwidth = 0;
std::map<std::pair<int, int>, int> Textout::busyStates;

inline void setCursorPos(const int line, const int column = 0)
{
  int y = line >= 0 ? line + 1 : 1;
  int x = column >= 0 ? column + 1 : 1;
  std::cout
    << std::string("\033[") + std::to_string(y) + ";" + std::to_string(x) + "H"
    << std::flush;
}

enum EraseMode
{
  FROM_CURSOR_TO_END_OF_LINE,
  FROM_START_OF_LINE_TO_CURSOR,
  ENTIRE_LINE,
  FROM_CURSOR_TO_END_OF_SCREEN,
  FROM_START_OF_SCREEN_TO_CURSOR,
  ENTIRE_SCREEN,
};
inline void erase(EraseMode mode)
{
  switch (mode)
  {
  case FROM_CURSOR_TO_END_OF_LINE:
    std::cout << "\033[0K" << std::flush;
    break;
  case FROM_START_OF_LINE_TO_CURSOR:
    std::cout << "\033[1K" << std::flush;
    break;
  case ENTIRE_LINE:
    std::cout << "\033[2K" << std::flush;
    break;
  case FROM_CURSOR_TO_END_OF_SCREEN:
    std::cout << "\033[0J" << std::flush;
    break;
  case FROM_START_OF_SCREEN_TO_CURSOR:
    std::cout << "\033[1J" << std::flush;
    break;
  case ENTIRE_SCREEN:
    std::cout << "\033[2J" << std::flush;
    break;
  }
}

std::pair<int, int> getConsoleSize() {
#ifdef _WIN32
  CONSOLE_SCREEN_BUFFER_INFO csbi;
  GetConsoleScreenBufferInfo(GetStdHandle(STD_OUTPUT_HANDLE), &csbi);
  return {(csbi.srWindow.Right - csbi.srWindow.Left + 1),
          (csbi.srWindow.Bottom - csbi.srWindow.Top + 1)};
#else
  struct winsize w;
  ioctl(STDOUT_FILENO, TIOCGWINSZ, &w);
  return {w.ws_col, w.ws_row};
#endif
}

void Textout::print(int line, int column, const std::string& s, bool nolinebreak)
{
#if TEXT_MODE == TEXT_MODE_NONE
  return;
#elif TEXT_MODE == TEXT_MODE_POSITIONED
  setCursorPos(curLine + line, column);
  erase(FROM_CURSOR_TO_END_OF_LINE);
  setCursorPos(curLine + line, column); // TODO: superfluous?
  std::cout << s << std::flush;
#elif TEXT_MODE == TEXT_MODE_FLAT
  std::cout << s << std::flush;
  if (!nolinebreak) std::cout << std::endl;
#endif
}

void Textout::printvar(int line, int column, const std::string& format, ...)
{
#if TEXT_MODE == TEXT_MODE_NONE
  return;
#endif
  va_list args;
  va_start( args, &format);
  vsprintf(text, format.c_str(), args);
  va_end(args);
  print(line, column, text, false);
}

void Textout::pagebreak()
{
#if TEXT_MODE == TEXT_MODE_NONE
  return;
#endif
  int height = getConsoleSize().second;
  throw "not yet implemented";
}

void Textout::printbreak(int line)
{
  for (int i = 0; i < bufferwidth; i ++)
  {
    text[i] = '-';
  }
  printvar(line, 0, "%s", text);
}

void Textout::printbusy(int line, int column)
{
#if TEXT_MODE == TEXT_MODE_POSITIONED
  switch (busyStates[std::pair<int,int>(line, column)])
  {
  case 0:
    print(line, column, "/");
    busyStates[std::pair<int,int>(line, column)] = 1;
    break;
  case 1:
    print(line, column, "-");
    busyStates[std::pair<int,int>(line, column)] = 2;
    break;
  case 2:
    print(line, column, "\\");
    busyStates[std::pair<int,int>(line, column)] = 3;
    break;
  case 3:
    print(line, column, "|");
    busyStates[std::pair<int,int>(line, column)] = 0;
    break;
  default:
    busyStates[std::pair<int,int>(line, column)] = 0;
    break;
  }
#endif
}

void Textout::init()
{
#if _WIN32
  // enable ANSI escape sequences (since Windows 10 TH2 (v1511))
  // see https://docs.microsoft.com/windows/console/console-virtual-terminal-sequences
  HANDLE hOut = GetStdHandle(STD_OUTPUT_HANDLE);
  if (hOut != INVALID_HANDLE_VALUE)
  {
    DWORD dwMode = 0;
    if (GetConsoleMode(hOut, &dwMode))
    {
      dwMode |= ENABLE_VIRTUAL_TERMINAL_PROCESSING;
      SetConsoleMode(hOut, dwMode | ENABLE_VIRTUAL_TERMINAL_PROCESSING);
    }
  }
#endif
#if TEXT_MODE == TEXT_MODE_NONE
  return;
#endif
  bufferwidth = getConsoleSize().first;
  clear();
}

void Textout::clear()
{
#if TEXT_MODE == TEXT_MODE_NONE
  return;
#endif
  erase(ENTIRE_SCREEN);
  setCursorPos(0, 0);
} 
