#pragma warning(disable:4244)
#pragma warning(disable:4018)

#include "definitions.h"
#include "textoutput.h"
#include "particle.h"
#include "tetramesh.h"

DclVector getWorld();
void  loadTetraFile(Mesh& m, char *filename);
void  computeSpringSet(Mesh& m);
void  updateParticles(Mesh& m);


const int MAX_NUM_PARTICLES = 10000;

// Integration step h
const DclFloat  TIMESTEP = 0.005;

// Elasticity of collision with walls
const DclFloat  WORLD_ELASTICITY = 1.0;

const DclFloat  SPRING_CONSTANT = 20000.0;
const DclFloat  SPRING_DAMPING_CONSTANT = 10.0;
const DclFloat  FRICTION = 100;

const Integration INTEGRATION = RungeKutta4;

//////////// GLOBALS /////////////////
std::vector<Mesh> vectorMeshes;

int frame = 0;

//////////////////////////////////////

void  visualizeParticles();

//------------------------------------------------------------------------------------------------------------------------------------

void  initializeTetraMesh(Mesh& m, DclVector translation)
{
  // initialize particles
  for (int p = 0; p < m.vectorParticles.size(); p ++)
  {
    Particle& particle = m.vectorParticles[p];
    particle.position += translation;
    particle.positionInit = particle.position;
    particle.mass = 1.0;
    particle.r = m.r; particle.g = m.g; particle.b = m.b;
  }
}

void  initializeRigidBody(Mesh& m)
{
  m.position.clear();
  m.mass = 0;

  // sum the masses of the particles to obtain the total rigid body mass and the center of mass
  for (int p = 0; p < m.vectorParticles.size(); p ++)
  {
    Particle& particle = m.vectorParticles[p];

    // TODO: Compute center of mass
    //m.position += ...
  }


  m.positionInit = m.position;

  // compute the inertia tensor
  m.I.clear();
  for (int p = 0; p < m.vectorParticles.size(); p ++)
  {  
    Particle& particle = m.vectorParticles[p];
    
    // TODO: Compute inertia tensor
    // m.I = m.I - ...
  }
  // inverted inertia tensor
  m.Iinv = m.I;
  m.Iinv.inverse();


  // obtain an arbitrary orthonormal orientation matrix

  // TODO: Compute an arbitrary initial orthonormal orientation matrix
  // m.orientation = DclMatrix33(x_axis, y_axis, z_axis);
}


//------------------------------------------------------------------------------------------------------------------------------------

// initialize the particle properties
void  initializeSimulation(int argc, char ** argv)
{
  std::string filename;
  if (argc > 1) 
  {
    filename = argv[1];
  }
  else
  {
    filename = "cube.dcstructure";
  }
  vectorMeshes.push_back(Mesh());

  Mesh& m1 = vectorMeshes[0];
  m1.integration = Euler;
  loadTetraFile(m1, (char*)filename.c_str());
  computeSpringSet(m1);
  initializeTetraMesh(m1, DclVector(0, 0, 0));

  initializeRigidBody(m1);
  updateParticles(m1);
}

//------------------------------------------------------------------------------------------------------------------------------------

// add external forces to the sum of forces acting on the particle
void  addForces(Mesh& m)
{
  m.force += DclVector(0, 1.0, 0) * m.mass * -9.81; // gravitational force
}

//------------------------------------------------------------------------------------------------------------------------------------

// perform one time-integration step
void  integrate(Mesh& m)
{
  // sum the forces and torques of all particles to obtain the total force and torque
  for (int i = 0; i < m.vectorParticles.size(); i ++)
  {
    Particle& particle = m.vectorParticles[i];

    // TODO: Compute the total force and torque on the rigid body
    // m.force += ...
    // m.torque += ...
  }

  // Euler-Cromer scheme
  
  // TODO: Euler-Cromer scheme for the center of mass (update m.position and m.velocity)

  // TODO: Numerically integrate the orientation of the rigid body
  // I.e. update the angular velocity and the orientation matrix of the rigid body
  // Try to avoid inverting the inertia tensor.

  // TODO: re-orthonormalize the orientation matrix
}

//------------------------------------------------------------------------------------------------------------------------------------

// collision handling with the world
void  worldCollision(Mesh &m, Particle& particle)
{
  DclVector world = getWorld();

  world *= 0.5;

  DclVector r = particle.position - m.position;
  DclVector penetrationDepth(0, 0, 0);

  // collision test with left and right wall
  if (particle.position.x < -world.x || particle.position.x > world.x)
  {
    penetrationDepth = DclVector(SIGN(particle.position.x) * world.x - particle.position.x, 0, 0);
  }

  // collision test with floor and ceiling
  if (particle.position.y < -world.y || particle.position.y > world.y)
  {
    penetrationDepth = DclVector(0, SIGN(particle.position.y) * world.y - particle.position.y, 0);
  }

  // collision test with front and back wall
  if (particle.position.z < -world.z || particle.position.z > world.z)
  {
    penetrationDepth = DclVector(0, 0, SIGN(particle.position.z) * world.z - particle.position.z);
  }

  // this mass-point is collided
  if (penetrationDepth.length2())
  {
    // the contact space (or contact normal) is d / |d|
    DclVector contactSpace(penetrationDepth); contactSpace.normalize();

    // TODO: compute the penalty force of this mass point
    // particle.force += ...

    // TODO: compute v_i, v_perp and v_t of the particle

    // TODO: compute damping force
    // particle.force -= ...

    // TODO: compute friction force
    // particle.force -= ...
  }
}

//------------------------------------------------------------------------------------------------------------------------------------

// compute and output the total energies in the scene
void  computeEnergies(Mesh& m, int i)
{
  DclFloat  e_kin = 0, e_pot = 0;
  DclVector world = getWorld();

  // sum the potential energy of each point
  for (int part = 0; part < m.vectorParticles.size(); part ++)
  {
    Particle& particle = m.vectorParticles[part];

    e_pot += (particle.position.y + world.z * 0.5) * particle.mass * 9.81;
  }

  // TODO: compute kinetic energy of cm
  
  DclFloat  e_rot = 0;
  // TODO: compute rotational energy


  Textout::printvar(6 + i, 0, "Ekin %f, Epot %f, Erot %f, Sum %f", e_kin, e_pot, e_rot, e_kin + e_pot + e_rot);
}


void  updateParticles(Mesh& m)
{
  // update the positions of each particle based on the orientation of the rigid body
  for (int part = 0; part < m.vectorParticles.size(); part ++)
  {
    DclVector d = m.vectorParticles[part].positionInit - m.positionInit;
    d = m.orientation * d;
    m.vectorParticles[part].position = d;
    m.vectorParticles[part].position += m.position;
  }

}

//------------------------------------------------------------------------------------------------------------------------------------

// perform one simulation step
void  simulationStep()
{
  static int splitCounter = 0;

  int numParticles = 0;

  for (int mesh = 0; mesh < vectorMeshes.size(); mesh ++)
  {
    Mesh& m = vectorMeshes[mesh];

    // clear the forces of the particles
    for (int part = 0; part < m.vectorParticles.size(); part ++)
    {
      m.vectorParticles[part].force.clear();
    }
    m.force.clear();
    m.torque.clear();

    // compute external forces on the particles
    addForces(m);

    // compute collision impulses
    for (int part = 0; part < m.vectorParticles.size(); part ++)
    {
      worldCollision(m, m.vectorParticles[part]);
    }

    integrate(m);

    // compute and output energies
    computeEnergies(m, mesh);
    // update the position of the particles
    updateParticles(m);

  }

  // draw particles
  visualizeParticles();

  // output statistics
  Textout::printvar(10, 0, "Frame %d, Num. particles %d", frame, numParticles);
  frame ++;
}


