/******************************************************************************
 * quaternion.inl
 * Copyright (C) 2004 Bruno Heidelberger <heidelberger@inf.ethz.ch>
 *                    Matthias Teschner <teschner@inf.ethz.ch>
 ******************************************************************************/

 /*****************************************************************************/
/** Constructs the quaternion instance.
  *
  * This function is the default constructor of the quaternion instance.
  *****************************************************************************/

inline DclQuaternion::DclQuaternion()
  : x(0), y(0), z(0), w((DclFloat)1)
{
}

 /*****************************************************************************/
/** Constructs the quaternion instance.
  *
  * This function is a constructor of the quaternion instance.
  *
  * @param q The quaternion to construct this quaternion instance from.
  *****************************************************************************/

inline DclQuaternion::DclQuaternion(const DclQuaternion& q)
  : x(q.x), y(q.y), z(q.z), w(q.w)
{
}

 /*****************************************************************************/
/** Constructs the quaternion instance.
  *
  * This function is a constructor of the quaternion instance.
  *
  * @param qx The x component.
  * @param qy The y component.
  * @param qz The z component.
  * @param qw The w component.
  *****************************************************************************/

inline DclQuaternion::DclQuaternion(DclFloat qx, DclFloat qy, DclFloat qz, DclFloat qw)
  : x(qx), y(qy), z(qz), w(qw)
{
}

 /*****************************************************************************/
/** Constructs the quaternion instance.
  *
  * This function is a constructor of the quaternion instance.
  *
  * @param v The axis component.
  * @param d The angle component.
  *****************************************************************************/

inline DclQuaternion::DclQuaternion(const DclVector& v, DclFloat d)
{
  set(v, d);
}


 /*****************************************************************************/

inline DclQuaternion::DclQuaternion (DclVector& e1, DclVector& e2, DclVector& e3)
{
  const DclFloat EPSILON = 0.0001;
  // Construction following [Shoemake, Animating Rotation with Quaternion Curves]
  DclFloat  omega2 = 0.25 * (1.0 + e1[0] + e2[1] + e3[2]);
  if (omega2 > EPSILON)
  {
    q4 = sqrt(omega2);
    //q4 = 0.5 * sqrt(1.0 + e1[0] + e2[1] + e3[2]);
    q1 = 1.0 / (4.0 * q4) * (e2[2] - e3[1]);
    q2 = 1.0 / (4.0 * q4) * (e3[0] - e1[2]);
    q3 = 1.0 / (4.0 * q4) * (e1[1] - e2[0]);
  }
  else
  {
    q4 = 0;
    DclFloat  x2 = -1.0/2.0 * (e2[1] + e3[2]);
    if (x2 > EPSILON)
    {
      q1 = sqrt(x2);
      q2 = e1[1] / (2.0 * q1);
      q3 = e1[2] / (2.0 * q1);
    }
    else
    {
      q1 = 0;
      DclFloat y2 = 1/2 * (1.0 - e3[2]);
      if (y2 > EPSILON)
      {
        q2 = sqrt(y2);
        q3 = e2[2] / (2.0 * q2);
      }
      else
      {
        q2 = q3 = 0;
      }
    }
  }
}

 /*****************************************************************************/
/** Destructs the quaternion instance.
  *
  * This function is the destructor of the quaternion instance.
  *****************************************************************************/

inline DclQuaternion::~DclQuaternion()
{
}

 /*****************************************************************************/
/** Interpolates the quaternion instance to another quaternion.
  *
  * This function interpolates the quaternion instance to another quaternion by
  * a given factor.
  *
  * @param d The blending factor in the range [0.0, 1.0].
  * @param v The quaternion to be interpolated to.
  *****************************************************************************/

inline void DclQuaternion::blend(DclFloat d, const DclQuaternion& q)
{
  DclFloat norm;
  norm = x * q.x + y * q.y + z * q.z + w * q.w;

  bool bFlip;
  bFlip = false;

  if(norm < 0)
  {
    norm = -norm;
    bFlip = true;
  }

  DclFloat inv_d;
  if((DclFloat)1.0 - norm < (DclFloat)0.000001)
  {
    inv_d = (DclFloat)1.0 - d;
  }
  else
  {
    DclFloat theta;
    theta = (DclFloat) acos(norm);

    DclFloat s;
    s = (DclFloat) (1.0 / sin(theta));

    inv_d = (DclFloat) sin((1.0 - d) * theta) * s;
    d = (DclFloat) sin(d * theta) * s;
  }

  if(bFlip)
  {
    d = -d;
  }

  x = inv_d * x + d * q.x;
  y = inv_d * y + d * q.y;
  z = inv_d * z + d * q.z;
  w = inv_d * w + d * q.w;
}

 /*****************************************************************************/
/** Clears the quaternion instance.
  *
  * This function clears the quaternion instance.
  *****************************************************************************/

inline void DclQuaternion::clear()
{
  x = 0;
  y = 0;
  z = 0;
  w = 1.0;
}

 /*****************************************************************************/
/** Conjugates the quaternion instance.
  *
  * This function conjugates the quaternion instance.
  *****************************************************************************/

inline void DclQuaternion::conjugate()
{
  x = -x;
  y = -y;
  z = -z;
}

 /*****************************************************************************/
/** Inverts the quaternion instance.
  *
  * This function inverts the quaternion instance.
  *****************************************************************************/

inline void DclQuaternion::invert()
{
  conjugate();
  normalize();
}

 /*****************************************************************************/

inline void DclQuaternion::normalize()
{
  const DclFloat norm = (x * x) + (y * y) + (z * z) + (w * w);

  if (norm == 0) return;
  
  const DclFloat inv_norm = 1 / sqrt(norm);
  x *= inv_norm;
  y *= inv_norm;
  z *= inv_norm;
  w *= inv_norm;  
}

inline void DclQuaternion::nullize()
{
  w = x = y = z = 0;
}

 /*****************************************************************************/
/** Sets new values.
  *
  * This function sets new values in the quaternion instance.
  *
  * @param v The axis component.
  * @param d The angle component.
  *****************************************************************************/

inline void DclQuaternion::set(const DclVector& v, DclFloat d)
{
  DclVector a(v);
  a.normalize();
  DclFloat f = sin(d / (DclFloat)2.0);
  x = a[0] * f;
  y = a[1] * f;
  z = a[2] * f;
  w = cos(d / (DclFloat)2.0);
}

 /*****************************************************************************/
/** Sets new values.
  *
  * This function sets new values in the quaternion instance.
  *
  * @param qx The x component.
  * @param qy The y component.
  * @param qz The z component.
  * @param qw The w component.
  *****************************************************************************/

inline void DclQuaternion::set(DclFloat qx, DclFloat qy, DclFloat qz, DclFloat qw)
{
  x = qx;
  y = qy;
  z = qz;
  w = qw;
}

 /*****************************************************************************/


inline void DclQuaternion::operator*=(const DclQuaternion& q)
{
  DclFloat qx, qy, qz, qw;
  qx = x;
  qy = y;
  qz = z;
  qw = w;

  x = qw * q.x + qx * q.w + qy * q.z - qz * q.y;
  y = qw * q.y - qx * q.z + qy * q.w + qz * q.x;
  z = qw * q.z + qx * q.y - qy * q.x + qz * q.w;
  w = qw * q.w - qx * q.x - qy * q.y - qz * q.z;
}

 /*****************************************************************************/


inline void DclQuaternion::operator+=(const DclQuaternion& dq)
{
  q1 += dq.q1;
  q2 += dq.q2;
  q3 += dq.q3;
  q4 += dq.q4;
}

 /*****************************************************************************/


inline void DclQuaternion::operator*=(const DclVector& v)
{
  DclFloat qx, qy, qz, qw;
  qx = x;
  qy = y;
  qz = z;
  qw = w;

  x = qw * v.x            + qy * v.z - qz * v.y;
  y = qw * v.y - qx * v.z            + qz * v.x;
  z = qw * v.z + qx * v.y - qy * v.x;
  w =          - qx * v.x - qy * v.y - qz * v.z;
}

/*****************************************************************************/

inline void DclQuaternion::operator*=(const DclFloat& f)
{
  x *= f; y *= f; z *= f; w *= f;
}

 /*****************************************************************************/
/** Provides access to the components of the quaternion instance.
  *
  * This function provides read and write access to the three components of the
  * quaternion instance.
  *
  * @param index The index to the specific component.
  *
  * @return A reference to the specific component.
  *****************************************************************************/

inline DclFloat& DclQuaternion::operator[](unsigned int i)
{
  return (&x)[i];
}

 /*****************************************************************************/
/** Provides access to the components of the quaternion instance.
  *
  * This function provides read access to the three components of the quaternion
  * instance.
  *
  * @param index The index to the specific component.
  *
  * @return A constant reference to the specific component.
  *****************************************************************************/

inline const DclFloat& DclQuaternion::operator[](unsigned int i) const
{
  return (&x)[i];
}

 /*****************************************************************************/
/** Equates the quaternion instance with another quaternion.
  *
  * This operator equates the quaternion instance with another quaternion.
  *
  * @param q The quaternion to equate the quaternion instance with.
  *****************************************************************************/

inline void DclQuaternion::operator=(const DclQuaternion& q)
{
  x = q.x;
  y = q.y;
  z = q.z;
  w = q.w;
}

 /*****************************************************************************/
/** Calculates the product of two quaternions.
  *
  * This operator calculates the product of two quaternions.
  *
  * @param q The first quaternion.
  * @param r The second quaternion.
  *
  * @return The product of the two quaternions.
  *****************************************************************************/

inline DclQuaternion operator*(const DclQuaternion& q, const DclQuaternion& r)
{
  return DclQuaternion(
    r.w * q.x + r.x * q.w + r.y * q.z - r.z * q.y,
    r.w * q.y - r.x * q.z + r.y * q.w + r.z * q.x,
    r.w * q.z + r.x * q.y - r.y * q.x + r.z * q.w,
    r.w * q.w - r.x * q.x - r.y * q.y - r.z * q.z
  );
}


inline DclQuaternion operator+(const DclQuaternion& q, const DclQuaternion& r)
{
  return DclQuaternion(q.q1 + r.q1, q.q2 + r.q2, q.q3 + r.q3, q.q4 + r.q4);
}

//****************************************************************************//

inline DclQuaternion operator*(const DclQuaternion& q, const DclFloat s)
{
  return DclQuaternion(q.q1 * s, q.q2 * s, q.q3 * s, q.q4 * s);
}

//****************************************************************************//

inline DclQuaternion DclQuaternion::multiply(const DclQuaternion& q)
{
  DclQuaternion result;

  multiply(result, q.q1, q.q2, q.q3, q.q4);

  return result;
}

inline void DclQuaternion::multiply(DclQuaternion& q_target, DclFloat qq1, DclFloat qq2, DclFloat qq3, DclFloat qq4)
{
  /*q_target.q1 = q1 * qq1 - q2 * qq2 - q3 * qq3 - q4 * qq4;
  q_target.q2 = q2 * qq1 + q1 * qq2 - q4 * qq3 + q3 * qq4;
  q_target.q3 = q3 * qq1 + q4 * qq2 + q1 * qq3 - q2 * qq4;
  q_target.q4 = q4 * qq1 - q3 * qq2 + q2 * qq3 + q1 * qq4;*/

  q_target.q1 =  q4 * qq1 - q3 * qq2 + q2 * qq3 + q1 * qq4;
  q_target.q2 =  q3 * qq1 + q4 * qq2 - q1 * qq3 + q2 * qq4;
  q_target.q3 = -q2 * qq1 + q1 * qq2 + q4 * qq3 + q3 * qq4;
  q_target.q4 = -q1 * qq1 - q2 * qq2 - q3 * qq3 + q4 * qq4;
}

//****************************************************************************//

inline DclQuaternion DclQuaternion::multiplyInverse(const DclQuaternion& q)
{
  DclQuaternion result;

  multiplyInverse(result, q.q1, q.q2, q.q3, q.q4);

  return result;
}

inline void DclQuaternion::multiplyInverse(DclQuaternion& q_target, DclFloat qq1, DclFloat qq2, DclFloat qq3, DclFloat qq4)
{
  /*q_target.q1 = q1 * qq1 - q2 * qq2 - q3 * qq3 - q4 * qq4;
  q_target.q2 = q2 * qq1 + q1 * qq2 + q4 * qq3 - q3 * qq4;
  q_target.q3 = q3 * qq1 - q4 * qq2 + q1 * qq3 + q2 * qq4;
  q_target.q4 = q4 * qq1 + q3 * qq2 - q2 * qq3 + q1 * qq4;*/

  q_target.q1 =  q4 * qq1 + q3 * qq2 - q2 * qq3 + q1 * qq4;
  q_target.q2 = -q3 * qq1 + q4 * qq2 + q1 * qq3 + q2 * qq4;
  q_target.q3 =  q2 * qq1 - q1 * qq2 + q4 * qq3 + q3 * qq4;
  q_target.q4 = -q1 * qq1 - q2 * qq2 - q3 * qq3 + q4 * qq4;
}

//****************************************************************************//

inline DclQuaternion DclQuaternion::multiplyTranspose(const DclQuaternion& q)
{

  DclQuaternion result;

  multiplyTranspose(result, q.q1, q.q2, q.q3, q.q4);

  return result;
}

inline void DclQuaternion::multiplyTranspose(DclQuaternion& q_target, DclFloat qq1, DclFloat qq2, DclFloat qq3, DclFloat qq4)
{
  /*
  q_target.q1 =  q1 * qq1 + q2 * qq2 + q3 * qq3 + q4 * qq4;
  q_target.q2 = -q2 * qq1 + q1 * qq2 + q4 * qq3 - q3 * qq4;
  q_target.q3 = -q3 * qq1 - q4 * qq2 + q1 * qq3 + q2 * qq4;
  q_target.q4 = -q4 * qq1 + q3 * qq2 - q2 * qq3 + q1 * qq4;*/

  q_target.q1 =  q4 * qq1 + q3 * qq2 - q2 * qq3 - q1 * qq4;
  q_target.q2 = -q3 * qq1 + q4 * qq2 + q1 * qq3 - q2 * qq4;
  q_target.q3 =  q2 * qq1 - q1 * qq2 + q4 * qq3 - q3 * qq4;
  q_target.q4 =  q1 * qq1 + q2 * qq2 + q3 * qq3 + q4 * qq4;
}

inline DclQuaternion DclQuaternion::multiplyInverseTranspose(const DclQuaternion& q)
{

  DclQuaternion result;

  multiplyInverseTranspose(result, q.q1, q.q2, q.q3, q.q4);

  return result;
}

inline void DclQuaternion::multiplyInverseTranspose(DclQuaternion& q_target, DclFloat qq1, DclFloat qq2, DclFloat qq3, DclFloat qq4)
{
  /*q_target.q1 =  q1 * qq1 + q2 * qq2 + q3 * qq3 + q4 * qq4;
  q_target.q2 = -q2 * qq1 + q1 * qq2 - q4 * qq3 + q3 * qq4;
  q_target.q3 = -q3 * qq1 + q4 * qq2 + q1 * qq3 - q2 * qq4;
  q_target.q4 = -q4 * qq1 - q3 * qq2 + q2 * qq3 + q1 * qq4;*/

  q_target.q1 =  q4 * qq1 - q3 * qq2 + q2 * qq3 - q1 * qq4;
  q_target.q2 =  q3 * qq1 + q4 * qq2 - q1 * qq3 - q2 * qq4;
  q_target.q3 = -q2 * qq1 + q1 * qq2 + q4 * qq3 - q3 * qq4;
  q_target.q4 =  q1 * qq1 + q2 * qq2 + q3 * qq3 + q4 * qq4;
}

//****************************************************************************//

inline void DclQuaternion::toDirector(DclVector& d1, DclVector& d2, DclVector& d3, DclFloat scale)
{
  d1.set(q1 * q1 - q2 * q2 - q3 * q3 + q4 * q4, 2.0 * (q1 * q2 + q3 * q4), 2.0 * (q1 * q3 - q2 * q4));
  d2.set(2.0 * (q1 * q2 - q3 * q4), -q1 * q1 + q2 * q2 - q3 * q3 + q4 * q4, 2.0 * (q2 * q3 + q1 * q4));
  d3.set(2.0 * (q1 * q3 + q2 * q4), 2.0 * (q2 * q3 - q1 * q4), -q1 * q1 - q2 * q2 + q3 * q3 + q4 * q4);

  d1.normalize(); d1 *= scale;
  d2.normalize(); d2 *= scale;
  d3.normalize(); d3 *= scale;
}

//****************************************************************************//

inline void DclQuaternion::toDirector(int i, DclVector &d, DclFloat scale)
{
  switch (i)
  {
  case 1:
    d.set(q1 * q1 - q2 * q2 - q3 * q3 + q4 * q4, 2.0 * (q1 * q2 + q3 * q4), 2.0 * (q1 * q3 - q2 * q4));
    break;
  case 2:
    d.set(2.0 * (q1 * q2 - q3 * q4), -q1 * q1 + q2 * q2 - q3 * q3 + q4 * q4, 2.0 * (q2 * q3 + q1 * q4));
    break;
  case 3:
    d.set(2.0 * (q1 * q3 + q2 * q4), 2.0 * (q2 * q3 - q1 * q4), -q1 * q1 - q2 * q2 + q3 * q3 + q4 * q4);
    break;
  default:
    d.set(0, 0, 0);
    break;
  }
  d.normalize(); d *= scale;

}

//****************************************************************************//

inline DclVector DclQuaternion::rotate(DclVector& v)
{
  DclVector d1, d2, d3;
  toDirector(d1, d2, d3);
  return DclVector(d1.x * v.x + d2.x * v.y + d3.x * v.z, d1.y * v.x + d2.y * v.y + d3.y * v.z, d1.z * v.x + d2.z * v.y + d3.z * v.z);
}


inline void DclQuaternion::inverseMatrix(DclMatrix& q_inverse)
{
  double 
    Q11 = q1, Q12 = -q2, Q13 = -q3, Q14 = -q4,
    Q21 = q2, Q22 = q1, Q23 = -q4, Q24 = q3,
    Q31 = q3, Q32 = q4, Q33 = q1, Q34 = -q2,
    Q41 = q4, Q42 = -q3, Q43 = q2, Q44 = q1;

  /*q_inverse.m[0][0] = (Q44 * Q22 * Q33 - Q43 * Q22 * Q34 + Q43 * Q32 * Q24 + Q42 * Q23 * Q34 - Q44 * Q32 * Q23 - Q42 * Q24 * Q33) / (Q11 * Q44 * Q22 * Q33 - Q11 * Q43 * Q22 * Q34 + Q11 * Q43 * Q32 * Q24 + Q11 * Q42 * Q23 * Q34 - Q11 * Q44 * Q32 * Q23 - Q11 * Q42 * Q24 * Q33 - Q41 * Q14 * Q22 * Q33 + Q41 * Q13 * Q22 * Q34 + Q22 * Q43 * Q31 * Q14 - Q22 * Q44 * Q31 * Q13 - Q43 * Q31 * Q12 * Q24 - Q41 * Q24 * Q32 * Q13 - Q42 * Q23 * Q31 * Q14 - Q42 * Q21 * Q13 * Q34 + Q44 * Q32 * Q21 * Q13 - Q44 * Q33 * Q21 * Q12 + Q43 * Q34 * Q21 * Q12 + Q44 * Q31 * Q12 * Q23 + Q41 * Q23 * Q32 * Q14 + Q41 * Q12 * Q24 * Q33 + Q42 * Q24 * Q31 * Q13 + Q42 * Q21 * Q14 * Q33 - Q43 * Q32 * Q21 * Q14 - Q41 * Q12 * Q23 * Q34);
  q_inverse.m[0][1] = -(Q42 * Q13 * Q34 - Q44 * Q32 * Q13 + Q44 * Q33 * Q12 - Q43 * Q34 * Q12 - Q42 * Q14 * Q33 + Q43 * Q32 * Q14) / (Q11 * Q44 * Q22 * Q33 - Q11 * Q43 * Q22 * Q34 + Q11 * Q43 * Q32 * Q24 + Q11 * Q42 * Q23 * Q34 - Q11 * Q44 * Q32 * Q23 - Q11 * Q42 * Q24 * Q33 - Q41 * Q14 * Q22 * Q33 + Q41 * Q13 * Q22 * Q34 + Q22 * Q43 * Q31 * Q14 - Q22 * Q44 * Q31 * Q13 - Q43 * Q31 * Q12 * Q24 - Q41 * Q24 * Q32 * Q13 - Q42 * Q23 * Q31 * Q14 - Q42 * Q21 * Q13 * Q34 + Q44 * Q32 * Q21 * Q13 - Q44 * Q33 * Q21 * Q12 + Q43 * Q34 * Q21 * Q12 + Q44 * Q31 * Q12 * Q23 + Q41 * Q23 * Q32 * Q14 + Q41 * Q12 * Q24 * Q33 + Q42 * Q24 * Q31 * Q13 + Q42 * Q21 * Q14 * Q33 - Q43 * Q32 * Q21 * Q14 - Q41 * Q12 * Q23 * Q34);
  q_inverse.m[0][2] = (Q22 * Q43 * Q14 - Q22 * Q44 * Q13 - Q43 * Q12 * Q24 - Q42 * Q23 * Q14 + Q44 * Q12 * Q23 + Q42 * Q24 * Q13) / (Q11 * Q44 * Q22 * Q33 - Q11 * Q43 * Q22 * Q34 + Q11 * Q43 * Q32 * Q24 + Q11 * Q42 * Q23 * Q34 - Q11 * Q44 * Q32 * Q23 - Q11 * Q42 * Q24 * Q33 - Q41 * Q14 * Q22 * Q33 + Q41 * Q13 * Q22 * Q34 + Q22 * Q43 * Q31 * Q14 - Q22 * Q44 * Q31 * Q13 - Q43 * Q31 * Q12 * Q24 - Q41 * Q24 * Q32 * Q13 - Q42 * Q23 * Q31 * Q14 - Q42 * Q21 * Q13 * Q34 + Q44 * Q32 * Q21 * Q13 - Q44 * Q33 * Q21 * Q12 + Q43 * Q34 * Q21 * Q12 + Q44 * Q31 * Q12 * Q23 + Q41 * Q23 * Q32 * Q14 + Q41 * Q12 * Q24 * Q33 + Q42 * Q24 * Q31 * Q13 + Q42 * Q21 * Q14 * Q33 - Q43 * Q32 * Q21 * Q14 - Q41 * Q12 * Q23 * Q34);
  q_inverse.m[0][3] = -(Q14 * Q22 * Q33 - Q13 * Q22 * Q34 + Q12 * Q23 * Q34 + Q24 * Q32 * Q13 - Q23 * Q32 * Q14 - Q12 * Q24 * Q33) / (Q11 * Q44 * Q22 * Q33 - Q11 * Q43 * Q22 * Q34 + Q11 * Q43 * Q32 * Q24 + Q11 * Q42 * Q23 * Q34 - Q11 * Q44 * Q32 * Q23 - Q11 * Q42 * Q24 * Q33 - Q41 * Q14 * Q22 * Q33 + Q41 * Q13 * Q22 * Q34 + Q22 * Q43 * Q31 * Q14 - Q22 * Q44 * Q31 * Q13 - Q43 * Q31 * Q12 * Q24 - Q41 * Q24 * Q32 * Q13 - Q42 * Q23 * Q31 * Q14 - Q42 * Q21 * Q13 * Q34 + Q44 * Q32 * Q21 * Q13 - Q44 * Q33 * Q21 * Q12 + Q43 * Q34 * Q21 * Q12 + Q44 * Q31 * Q12 * Q23 + Q41 * Q23 * Q32 * Q14 + Q41 * Q12 * Q24 * Q33 + Q42 * Q24 * Q31 * Q13 + Q42 * Q21 * Q14 * Q33 - Q43 * Q32 * Q21 * Q14 - Q41 * Q12 * Q23 * Q34);
  q_inverse.m[1][0] = -(Q41 * Q23 * Q34 - Q23 * Q44 * Q31 + Q24 * Q43 * Q31 - Q41 * Q24 * Q33 + Q21 * Q44 * Q33 - Q21 * Q43 * Q34) / (Q11 * Q44 * Q22 * Q33 - Q11 * Q43 * Q22 * Q34 + Q11 * Q43 * Q32 * Q24 + Q11 * Q42 * Q23 * Q34 - Q11 * Q44 * Q32 * Q23 - Q11 * Q42 * Q24 * Q33 - Q41 * Q14 * Q22 * Q33 + Q41 * Q13 * Q22 * Q34 + Q22 * Q43 * Q31 * Q14 - Q22 * Q44 * Q31 * Q13 - Q43 * Q31 * Q12 * Q24 - Q41 * Q24 * Q32 * Q13 - Q42 * Q23 * Q31 * Q14 - Q42 * Q21 * Q13 * Q34 + Q44 * Q32 * Q21 * Q13 - Q44 * Q33 * Q21 * Q12 + Q43 * Q34 * Q21 * Q12 + Q44 * Q31 * Q12 * Q23 + Q41 * Q23 * Q32 * Q14 + Q41 * Q12 * Q24 * Q33 + Q42 * Q24 * Q31 * Q13 + Q42 * Q21 * Q14 * Q33 - Q43 * Q32 * Q21 * Q14 - Q41 * Q12 * Q23 * Q34);
  q_inverse.m[1][1] = (Q11 * Q44 * Q33 - Q11 * Q43 * Q34 + Q43 * Q31 * Q14 - Q44 * Q31 * Q13 - Q41 * Q14 * Q33 + Q41 * Q13 * Q34) / (Q11 * Q44 * Q22 * Q33 - Q11 * Q43 * Q22 * Q34 + Q11 * Q43 * Q32 * Q24 + Q11 * Q42 * Q23 * Q34 - Q11 * Q44 * Q32 * Q23 - Q11 * Q42 * Q24 * Q33 - Q41 * Q14 * Q22 * Q33 + Q41 * Q13 * Q22 * Q34 + Q22 * Q43 * Q31 * Q14 - Q22 * Q44 * Q31 * Q13 - Q43 * Q31 * Q12 * Q24 - Q41 * Q24 * Q32 * Q13 - Q42 * Q23 * Q31 * Q14 - Q42 * Q21 * Q13 * Q34 + Q44 * Q32 * Q21 * Q13 - Q44 * Q33 * Q21 * Q12 + Q43 * Q34 * Q21 * Q12 + Q44 * Q31 * Q12 * Q23 + Q41 * Q23 * Q32 * Q14 + Q41 * Q12 * Q24 * Q33 + Q42 * Q24 * Q31 * Q13 + Q42 * Q21 * Q14 * Q33 - Q43 * Q32 * Q21 * Q14 - Q41 * Q12 * Q23 * Q34);
  q_inverse.m[1][2] = -(-Q11 * Q43 * Q24 + Q11 * Q44 * Q23 + Q41 * Q24 * Q13 - Q41 * Q23 * Q14 - Q21 * Q13 * Q44 + Q21 * Q14 * Q43) / (Q11 * Q44 * Q22 * Q33 - Q11 * Q43 * Q22 * Q34 + Q11 * Q43 * Q32 * Q24 + Q11 * Q42 * Q23 * Q34 - Q11 * Q44 * Q32 * Q23 - Q11 * Q42 * Q24 * Q33 - Q41 * Q14 * Q22 * Q33 + Q41 * Q13 * Q22 * Q34 + Q22 * Q43 * Q31 * Q14 - Q22 * Q44 * Q31 * Q13 - Q43 * Q31 * Q12 * Q24 - Q41 * Q24 * Q32 * Q13 - Q42 * Q23 * Q31 * Q14 - Q42 * Q21 * Q13 * Q34 + Q44 * Q32 * Q21 * Q13 - Q44 * Q33 * Q21 * Q12 + Q43 * Q34 * Q21 * Q12 + Q44 * Q31 * Q12 * Q23 + Q41 * Q23 * Q32 * Q14 + Q41 * Q12 * Q24 * Q33 + Q42 * Q24 * Q31 * Q13 + Q42 * Q21 * Q14 * Q33 - Q43 * Q32 * Q21 * Q14 - Q41 * Q12 * Q23 * Q34);
  q_inverse.m[1][3] = (Q11 * Q23 * Q34 - Q11 * Q24 * Q33 + Q24 * Q31 * Q13 - Q23 * Q31 * Q14 - Q21 * Q13 * Q34 + Q21 * Q14 * Q33) / (Q11 * Q44 * Q22 * Q33 - Q11 * Q43 * Q22 * Q34 + Q11 * Q43 * Q32 * Q24 + Q11 * Q42 * Q23 * Q34 - Q11 * Q44 * Q32 * Q23 - Q11 * Q42 * Q24 * Q33 - Q41 * Q14 * Q22 * Q33 + Q41 * Q13 * Q22 * Q34 + Q22 * Q43 * Q31 * Q14 - Q22 * Q44 * Q31 * Q13 - Q43 * Q31 * Q12 * Q24 - Q41 * Q24 * Q32 * Q13 - Q42 * Q23 * Q31 * Q14 - Q42 * Q21 * Q13 * Q34 + Q44 * Q32 * Q21 * Q13 - Q44 * Q33 * Q21 * Q12 + Q43 * Q34 * Q21 * Q12 + Q44 * Q31 * Q12 * Q23 + Q41 * Q23 * Q32 * Q14 + Q41 * Q12 * Q24 * Q33 + Q42 * Q24 * Q31 * Q13 + Q42 * Q21 * Q14 * Q33 - Q43 * Q32 * Q21 * Q14 - Q41 * Q12 * Q23 * Q34);
  q_inverse.m[2][0] = (Q41 * Q22 * Q34 - Q22 * Q44 * Q31 - Q41 * Q24 * Q32 - Q42 * Q21 * Q34 + Q44 * Q32 * Q21 + Q42 * Q24 * Q31) / (Q11 * Q44 * Q22 * Q33 - Q11 * Q43 * Q22 * Q34 + Q11 * Q43 * Q32 * Q24 + Q11 * Q42 * Q23 * Q34 - Q11 * Q44 * Q32 * Q23 - Q11 * Q42 * Q24 * Q33 - Q41 * Q14 * Q22 * Q33 + Q41 * Q13 * Q22 * Q34 + Q22 * Q43 * Q31 * Q14 - Q22 * Q44 * Q31 * Q13 - Q43 * Q31 * Q12 * Q24 - Q41 * Q24 * Q32 * Q13 - Q42 * Q23 * Q31 * Q14 - Q42 * Q21 * Q13 * Q34 + Q44 * Q32 * Q21 * Q13 - Q44 * Q33 * Q21 * Q12 + Q43 * Q34 * Q21 * Q12 + Q44 * Q31 * Q12 * Q23 + Q41 * Q23 * Q32 * Q14 + Q41 * Q12 * Q24 * Q33 + Q42 * Q24 * Q31 * Q13 + Q42 * Q21 * Q14 * Q33 - Q43 * Q32 * Q21 * Q14 - Q41 * Q12 * Q23 * Q34);
  q_inverse.m[2][1] = -(Q11 * Q44 * Q32 - Q11 * Q42 * Q34 - Q44 * Q31 * Q12 + Q42 * Q31 * Q14 + Q41 * Q12 * Q34 - Q41 * Q32 * Q14) / (Q11 * Q44 * Q22 * Q33 - Q11 * Q43 * Q22 * Q34 + Q11 * Q43 * Q32 * Q24 + Q11 * Q42 * Q23 * Q34 - Q11 * Q44 * Q32 * Q23 - Q11 * Q42 * Q24 * Q33 - Q41 * Q14 * Q22 * Q33 + Q41 * Q13 * Q22 * Q34 + Q22 * Q43 * Q31 * Q14 - Q22 * Q44 * Q31 * Q13 - Q43 * Q31 * Q12 * Q24 - Q41 * Q24 * Q32 * Q13 - Q42 * Q23 * Q31 * Q14 - Q42 * Q21 * Q13 * Q34 + Q44 * Q32 * Q21 * Q13 - Q44 * Q33 * Q21 * Q12 + Q43 * Q34 * Q21 * Q12 + Q44 * Q31 * Q12 * Q23 + Q41 * Q23 * Q32 * Q14 + Q41 * Q12 * Q24 * Q33 + Q42 * Q24 * Q31 * Q13 + Q42 * Q21 * Q14 * Q33 - Q43 * Q32 * Q21 * Q14 - Q41 * Q12 * Q23 * Q34);
  q_inverse.m[2][2] = (Q44 * Q11 * Q22 - Q44 * Q21 * Q12 - Q41 * Q14 * Q22 - Q42 * Q11 * Q24 + Q42 * Q21 * Q14 + Q41 * Q12 * Q24) / (Q11 * Q44 * Q22 * Q33 - Q11 * Q43 * Q22 * Q34 + Q11 * Q43 * Q32 * Q24 + Q11 * Q42 * Q23 * Q34 - Q11 * Q44 * Q32 * Q23 - Q11 * Q42 * Q24 * Q33 - Q41 * Q14 * Q22 * Q33 + Q41 * Q13 * Q22 * Q34 + Q22 * Q43 * Q31 * Q14 - Q22 * Q44 * Q31 * Q13 - Q43 * Q31 * Q12 * Q24 - Q41 * Q24 * Q32 * Q13 - Q42 * Q23 * Q31 * Q14 - Q42 * Q21 * Q13 * Q34 + Q44 * Q32 * Q21 * Q13 - Q44 * Q33 * Q21 * Q12 + Q43 * Q34 * Q21 * Q12 + Q44 * Q31 * Q12 * Q23 + Q41 * Q23 * Q32 * Q14 + Q41 * Q12 * Q24 * Q33 + Q42 * Q24 * Q31 * Q13 + Q42 * Q21 * Q14 * Q33 - Q43 * Q32 * Q21 * Q14 - Q41 * Q12 * Q23 * Q34);
  q_inverse.m[2][3] = -(Q34 * Q11 * Q22 - Q34 * Q21 * Q12 - Q31 * Q14 * Q22 - Q32 * Q11 * Q24 + Q32 * Q21 * Q14 + Q31 * Q12 * Q24) / (Q11 * Q44 * Q22 * Q33 - Q11 * Q43 * Q22 * Q34 + Q11 * Q43 * Q32 * Q24 + Q11 * Q42 * Q23 * Q34 - Q11 * Q44 * Q32 * Q23 - Q11 * Q42 * Q24 * Q33 - Q41 * Q14 * Q22 * Q33 + Q41 * Q13 * Q22 * Q34 + Q22 * Q43 * Q31 * Q14 - Q22 * Q44 * Q31 * Q13 - Q43 * Q31 * Q12 * Q24 - Q41 * Q24 * Q32 * Q13 - Q42 * Q23 * Q31 * Q14 - Q42 * Q21 * Q13 * Q34 + Q44 * Q32 * Q21 * Q13 - Q44 * Q33 * Q21 * Q12 + Q43 * Q34 * Q21 * Q12 + Q44 * Q31 * Q12 * Q23 + Q41 * Q23 * Q32 * Q14 + Q41 * Q12 * Q24 * Q33 + Q42 * Q24 * Q31 * Q13 + Q42 * Q21 * Q14 * Q33 - Q43 * Q32 * Q21 * Q14 - Q41 * Q12 * Q23 * Q34);
  q_inverse.m[3][0] = -(-Q22 * Q43 * Q31 + Q41 * Q22 * Q33 + Q43 * Q32 * Q21 + Q42 * Q23 * Q31 - Q41 * Q23 * Q32 - Q42 * Q21 * Q33) / (Q11 * Q44 * Q22 * Q33 - Q11 * Q43 * Q22 * Q34 + Q11 * Q43 * Q32 * Q24 + Q11 * Q42 * Q23 * Q34 - Q11 * Q44 * Q32 * Q23 - Q11 * Q42 * Q24 * Q33 - Q41 * Q14 * Q22 * Q33 + Q41 * Q13 * Q22 * Q34 + Q22 * Q43 * Q31 * Q14 - Q22 * Q44 * Q31 * Q13 - Q43 * Q31 * Q12 * Q24 - Q41 * Q24 * Q32 * Q13 - Q42 * Q23 * Q31 * Q14 - Q42 * Q21 * Q13 * Q34 + Q44 * Q32 * Q21 * Q13 - Q44 * Q33 * Q21 * Q12 + Q43 * Q34 * Q21 * Q12 + Q44 * Q31 * Q12 * Q23 + Q41 * Q23 * Q32 * Q14 + Q41 * Q12 * Q24 * Q33 + Q42 * Q24 * Q31 * Q13 + Q42 * Q21 * Q14 * Q33 - Q43 * Q32 * Q21 * Q14 - Q41 * Q12 * Q23 * Q34);
  q_inverse.m[3][1] = (Q11 * Q43 * Q32 - Q11 * Q42 * Q33 - Q43 * Q31 * Q12 - Q41 * Q32 * Q13 + Q41 * Q12 * Q33 + Q42 * Q31 * Q13) / (Q11 * Q44 * Q22 * Q33 - Q11 * Q43 * Q22 * Q34 + Q11 * Q43 * Q32 * Q24 + Q11 * Q42 * Q23 * Q34 - Q11 * Q44 * Q32 * Q23 - Q11 * Q42 * Q24 * Q33 - Q41 * Q14 * Q22 * Q33 + Q41 * Q13 * Q22 * Q34 + Q22 * Q43 * Q31 * Q14 - Q22 * Q44 * Q31 * Q13 - Q43 * Q31 * Q12 * Q24 - Q41 * Q24 * Q32 * Q13 - Q42 * Q23 * Q31 * Q14 - Q42 * Q21 * Q13 * Q34 + Q44 * Q32 * Q21 * Q13 - Q44 * Q33 * Q21 * Q12 + Q43 * Q34 * Q21 * Q12 + Q44 * Q31 * Q12 * Q23 + Q41 * Q23 * Q32 * Q14 + Q41 * Q12 * Q24 * Q33 + Q42 * Q24 * Q31 * Q13 + Q42 * Q21 * Q14 * Q33 - Q43 * Q32 * Q21 * Q14 - Q41 * Q12 * Q23 * Q34);
  q_inverse.m[3][2] = -(Q43 * Q11 * Q22 - Q43 * Q21 * Q12 - Q41 * Q13 * Q22 - Q42 * Q11 * Q23 + Q42 * Q21 * Q13 + Q41 * Q12 * Q23) / (Q11 * Q44 * Q22 * Q33 - Q11 * Q43 * Q22 * Q34 + Q11 * Q43 * Q32 * Q24 + Q11 * Q42 * Q23 * Q34 - Q11 * Q44 * Q32 * Q23 - Q11 * Q42 * Q24 * Q33 - Q41 * Q14 * Q22 * Q33 + Q41 * Q13 * Q22 * Q34 + Q22 * Q43 * Q31 * Q14 - Q22 * Q44 * Q31 * Q13 - Q43 * Q31 * Q12 * Q24 - Q41 * Q24 * Q32 * Q13 - Q42 * Q23 * Q31 * Q14 - Q42 * Q21 * Q13 * Q34 + Q44 * Q32 * Q21 * Q13 - Q44 * Q33 * Q21 * Q12 + Q43 * Q34 * Q21 * Q12 + Q44 * Q31 * Q12 * Q23 + Q41 * Q23 * Q32 * Q14 + Q41 * Q12 * Q24 * Q33 + Q42 * Q24 * Q31 * Q13 + Q42 * Q21 * Q14 * Q33 - Q43 * Q32 * Q21 * Q14 - Q41 * Q12 * Q23 * Q34);
  q_inverse.m[3][3] = 0.1e1 / (Q11 * Q44 * Q22 * Q33 - Q11 * Q43 * Q22 * Q34 + Q11 * Q43 * Q32 * Q24 + Q11 * Q42 * Q23 * Q34 - Q11 * Q44 * Q32 * Q23 - Q11 * Q42 * Q24 * Q33 - Q41 * Q14 * Q22 * Q33 + Q41 * Q13 * Q22 * Q34 + Q22 * Q43 * Q31 * Q14 - Q22 * Q44 * Q31 * Q13 - Q43 * Q31 * Q12 * Q24 - Q41 * Q24 * Q32 * Q13 - Q42 * Q23 * Q31 * Q14 - Q42 * Q21 * Q13 * Q34 + Q44 * Q32 * Q21 * Q13 - Q44 * Q33 * Q21 * Q12 + Q43 * Q34 * Q21 * Q12 + Q44 * Q31 * Q12 * Q23 + Q41 * Q23 * Q32 * Q14 + Q41 * Q12 * Q24 * Q33 + Q42 * Q24 * Q31 * Q13 + Q42 * Q21 * Q14 * Q33 - Q43 * Q32 * Q21 * Q14 - Q41 * Q12 * Q23 * Q34) * (Q33 * Q11 * Q22 - Q33 * Q21 * Q12 - Q31 * Q13 * Q22 - Q32 * Q11 * Q23 + Q32 * Q21 * Q13 + Q31 * Q12 * Q23);
*/
  // TODO: Correct
  exit(0);
}